# OBSS SAHI Tool
# Code written by Fatih C Akyon, 2020.

import os
import time

from tqdm import tqdm

from sahi.postprocess.match import PredictionMatcher
from sahi.postprocess.merge import PredictionMerger, ScoreMergingPolicy
from sahi.postprocess.ops import box_intersection, box_ios, box_union
from sahi.prediction import ObjectPrediction, PredictionInput
from sahi.slicing import slice_image
from sahi.utils.coco import Coco
from sahi.utils.cv import (
    crop_object_predictions,
    read_image,
    visualize_object_predictions,
)
from sahi.utils.file import (
    Path,
    get_base_filename,
    import_class,
    increment_path,
    list_files,
    save_json,
    save_pickle,
)
from sahi.utils.torch import to_float_tensor


def get_prediction(
    image,
    detection_model,
    shift_amount: list = [0, 0],
    full_shape=None,
    merger=None,
    matcher=None,
    verbose: int = 0,
):
    """
    Function for performing prediction for given image using given detection_model.

    Arguments:
        image: str or np.ndarray
            Location of image or numpy image matrix to slice
        detection_model: model.DetectionMode
        shift_amount: List
            To shift the box and mask predictions from sliced image to full
            sized image, should be in the form of [shift_x, shift_y]
        full_shape: List
            Size of the full image, should be in the form of [height, width]
        merger: postprocess.PredictionMerger
        matcher: postprocess.PredictionMatcher
        verbose: int
            0: no print (default)
            1: print prediction duration

    Returns:
        A dict with fields:
            object_prediction_list: a list of ObjectPrediction
            durations_in_seconds: a dict containing elapsed times for profiling
    """
    durations_in_seconds = dict()

    # read image if image is str
    if isinstance(image, str):
        image = read_image(image)
    # get prediction
    time_start = time.time()
    detection_model.perform_inference(image)
    time_end = time.time() - time_start
    durations_in_seconds["prediction"] = time_end

    # process prediction
    time_start = time.time()
    # works only with 1 batch
    detection_model.convert_original_predictions(
        shift_amount=shift_amount,
        full_shape=full_shape,
    )
    object_prediction_list = detection_model.object_prediction_list
    # filter out predictions with lower score
    filtered_object_prediction_list = [
        object_prediction
        for object_prediction in object_prediction_list
        if object_prediction.score.score > detection_model.prediction_score_threshold
    ]
    # merge matching predictions
    if merger is not None:
        filtered_object_prediction_list = merger.merge_batch(
            matcher,
            filtered_object_prediction_list,
            merge_type="merge",
        )
    else:
        # init match merge instances
        merger = PredictionMerger(
            score_merging=ScoreMergingPolicy.LARGER_SCORE, box_merger=box_union
        )
        matcher = PredictionMatcher(threshold=0.5, scorer=box_ios)
        # merge matching predictions
        filtered_object_prediction_list = merger.merge_batch(
            matcher,
            filtered_object_prediction_list,
            merge_type="merge",
            ignore_class_label=True,
        )

    time_end = time.time() - time_start
    durations_in_seconds["postprocess"] = time_end

    if verbose == 1:
        print(
            "Prediction performed in",
            durations_in_seconds["prediction"],
            "seconds.",
        )

    return {
        "object_prediction_list": filtered_object_prediction_list,
        "durations_in_seconds": durations_in_seconds,
    }


def get_sliced_prediction(
    image,
    detection_model=None,
    slice_height: int = 256,
    slice_width: int = 256,
    overlap_height_ratio: float = 0.2,
    overlap_width_ratio: float = 0.2,
    match_iou_threshold: float = 0.5,
    verbose: int = 1,
):
    """
    Function for slice image + get predicion for each slice + combine predictions in full image.

    Args:
        image: str or np.ndarray
            Location of image or numpy image matrix to slice
        detection_model: model.DetectionModel
        slice_height: int
            Height of each slice.  Defaults to ``256``.
        slice_width: int
            Width of each slice.  Defaults to ``256``.
        overlap_height_ratio: float
            Fractional overlap in height of each window (e.g. an overlap of 0.2 for a window
            of size 256 yields an overlap of 51 pixels).
            Default to ``0.2``.
        overlap_width_ratio: float
            Fractional overlap in width of each window (e.g. an overlap of 0.2 for a window
            of size 256 yields an overlap of 51 pixels).
            Default to ``0.2``.
        match_iou_threshold: float
            Sliced predictions having higher iou than match_iou_threshold will be merged.
        verbose: int
            0: no print
            1: print number of slices (default)
            2: print number of slices and slice/prediction durations

    Returns:
        A Dict with fields:
            object_prediction_list: a list of sahi.prediction.ObjectPrediction
            durations_in_seconds: a dict containing elapsed times for profiling
    """
    # for profiling
    durations_in_seconds = dict()

    # currently only 1 batch supported
    num_batch = 1

    # create slices from full image
    time_start = time.time()
    slice_image_result, num_total_invalid_segmentation = slice_image(
        image=image,
        slice_height=slice_height,
        slice_width=slice_width,
        overlap_height_ratio=overlap_height_ratio,
        overlap_width_ratio=overlap_width_ratio,
    )
    num_slices = len(slice_image_result)
    time_end = time.time() - time_start
    durations_in_seconds["slice"] = time_end

    # init match merge instances
    merger = PredictionMerger(
        score_merging=ScoreMergingPolicy.LARGER_SCORE, box_merger=box_union
    )
    matcher = PredictionMatcher(threshold=match_iou_threshold, scorer=box_ios)

    # create prediction input
    num_group = int(num_slices / num_batch)
    if verbose == 1 or verbose == 2:
        if num_slices > 0:
            print("Number of slices:", num_slices)
        else:
            print("Number of slices:", 1)
    object_prediction_list = []
    if num_slices > 0:  # if zero_frac < max_allowed_zeros_ratio from slice_image
        for group_ind in range(num_group):
            # prepare batch (currently supports only 1 batch)
            image_list = []
            shift_amount_list = []
            for image_ind in range(num_batch):
                image_list.append(
                    slice_image_result.images[group_ind * num_batch + image_ind]
                )
                shift_amount_list.append(
                    slice_image_result.starting_pixels[
                        group_ind * num_batch + image_ind
                    ]
                )
            # perform batch prediction
            prediction_result = get_prediction(
                image=image_list[0],
                detection_model=detection_model,
                shift_amount=shift_amount_list[0],
                full_shape=[
                    slice_image_result.original_image_height,
                    slice_image_result.original_image_width,
                ],
            )
            object_prediction_list.extend(prediction_result["object_prediction_list"])
    else:  # if zero_frac >= max_allowed_zeros_ratio from slice_image
        prediction_result = get_prediction(
            image=image,
            detection_model=detection_model,
            shift_amount=[0, 0],
            full_shape=None,
            merger=None,
            matcher=None,
            verbose=0,
        )
        object_prediction_list.extend(prediction_result["object_prediction_list"])

    # remove empty predictions
    object_prediction_list = [
        object_prediction
        for object_prediction in object_prediction_list
        if object_prediction
    ]
    # convert sliced predictions to full predictions
    full_object_prediction_list = []
    for object_prediction in object_prediction_list:
        full_object_prediction = object_prediction.get_shifted_object_prediction()
        full_object_prediction_list.append(full_object_prediction)

    time_end = time.time() - time_start
    durations_in_seconds["prediction"] = time_end

    if verbose == 2:
        print(
            "Slicing performed in",
            durations_in_seconds["slice"],
            "seconds.",
        )
        print(
            "Prediction performed in",
            durations_in_seconds["prediction"],
            "seconds.",
        )

    # merge matching predictions
    full_object_prediction_list = merger.merge_batch(
        matcher,
        full_object_prediction_list,
        merge_type="merge",
    )

    # return filtered elements of filtered full_object_prediction_list
    return {
        "object_prediction_list": full_object_prediction_list,
        "durations_in_seconds": durations_in_seconds,
    }


def predict(
    model_name="MmdetDetectionModel",
    model_parameters=None,
    source=None,
    apply_sliced_prediction: bool = True,
    slice_height: int = 256,
    slice_width: int = 256,
    overlap_height_ratio: float = 0.2,
    overlap_width_ratio: float = 0.2,
    match_iou_threshold: float = 0.5,
    export_visual=True,
    export_pickle=False,
    export_crop=False,
    coco_file_path=None,
    project="runs/predict",
    name="exp",
    visual_bbox_thickness: int = 1,
    visual_text_size: float = 0.3,
    visual_text_thickness: int = 1,
    visual_export_format: str = "png",
    verbose: int = 1,
):
    """
    Performs prediction for all present images in given folder.

    Args:
        model_name: str
            Name of the implemented DetectionModel in model.py file.
        model_parameter: a dict with fields:
            model_path: str
                Path for the instance segmentation model weight
            config_path: str
                Path for the mmdetection instance segmentation model config file
            prediction_score_threshold: float
                All predictions with score < prediction_score_threshold will be discarded.
            device: str
                Torch device, "cpu" or "cuda"
            category_remapping: dict: str to int
                Remap category ids after performing inference
        source: str
            Folder directory that contains images or path of the image to be predicted.
        apply_sliced_prediction: bool
            Set to True if you want sliced prediction, set to False for full prediction.
        slice_height: int
            Height of each slice.  Defaults to ``256``.
        slice_width: int
            Width of each slice.  Defaults to ``256``.
        overlap_height_ratio: float
            Fractional overlap in height of each window (e.g. an overlap of 0.2 for a window
            of size 256 yields an overlap of 51 pixels).
            Default to ``0.2``.
        overlap_width_ratio: float
            Fractional overlap in width of each window (e.g. an overlap of 0.2 for a window
            of size 256 yields an overlap of 51 pixels).
            Default to ``0.2``.
        match_iou_threshold: float
            Sliced predictions having higher iou than match_iou_threshold will be merged.
        export_pickle: bool
            Export predictions as .pickle
        export_crop: bool
            Export predictions as cropped images.
        coco_file_path: str
            If coco file path is provided, detection results will be exported in coco json format.
        project: str
            Save results to project/name.
        name: str
            Save results to project/name.
        visual_bbox_thickness: int
        visual_text_size: float
        visual_text_thickness: int
        visual_export_format: str
            Can be specified as 'jpg' or 'png'
        verbose: int
            0: no print
            1: print slice/prediction durations, number of slices, model loading/file exporting durations
    """
    # for profiling
    durations_in_seconds = dict()

    # list image files in directory
    if coco_file_path:
        coco = Coco.from_coco_dict_or_path(coco_file_path)
        image_path_list = [
            str(Path(source) / Path(coco_image.file_name)) for coco_image in coco.images
        ]
        image_id_list = [coco_image.id for coco_image in coco.images]
        coco_json = []
    elif os.path.isdir(source):
        time_start = time.time()
        image_path_list = list_files(
            directory=source,
            contains=[".jpg", ".jpeg", ".png"],
            verbose=verbose,
        )
        time_end = time.time() - time_start
        durations_in_seconds["list_files"] = time_end
    else:
        image_path_list = [source]
        durations_in_seconds["list_files"] = 0

    # init export directories
    save_dir = Path(
        increment_path(Path(project) / name, exist_ok=False)
    )  # increment run
    crop_dir = save_dir / "crops"
    visual_dir = save_dir / "visuals"
    pickle_dir = save_dir / "pickles"
    save_dir.mkdir(parents=True, exist_ok=True)  # make dir

    # init model instance
    time_start = time.time()
    DetectionModel = import_class(model_name)
    detection_model = DetectionModel(
        model_path=model_parameters["model_path"],
        config_path=model_parameters["config_path"],
        prediction_score_threshold=model_parameters["prediction_score_threshold"],
        device=model_parameters["device"],
        category_mapping=model_parameters["category_mapping"],
        category_remapping=model_parameters["category_remapping"],
        load_at_init=False,
    )
    detection_model.load_model()
    time_end = time.time() - time_start
    durations_in_seconds["model_load"] = time_end

    # iterate over source images
    durations_in_seconds["prediction"] = 0
    durations_in_seconds["slice"] = 0
    for ind, image_path in enumerate(tqdm(image_path_list)):
        # get filename
        filename_without_extension = str(Path(image_path).stem)
        # load image
        image = read_image(image_path)

        # perform prediction
        if apply_sliced_prediction:
            # get sliced prediction
            prediction_result = get_sliced_prediction(
                image=image,
                detection_model=detection_model,
                slice_height=slice_height,
                slice_width=slice_width,
                overlap_height_ratio=overlap_height_ratio,
                overlap_width_ratio=overlap_width_ratio,
                match_iou_threshold=match_iou_threshold,
                verbose=verbose,
            )
            object_prediction_list = prediction_result["object_prediction_list"]
            durations_in_seconds["slice"] += prediction_result["durations_in_seconds"][
                "slice"
            ]
        else:
            # get full sized prediction
            prediction_result = get_prediction(
                image=image,
                detection_model=detection_model,
                shift_amount=[0, 0],
                full_shape=None,
                merger=None,
                matcher=None,
                verbose=0,
            )
            object_prediction_list = prediction_result["object_prediction_list"]

        durations_in_seconds["prediction"] += prediction_result["durations_in_seconds"][
            "prediction"
        ]

        if coco_file_path:
            image_id = image_id_list[ind]
            for object_prediction in object_prediction_list:
                coco_prediction = object_prediction.to_coco_prediction()
                coco_prediction.image_id = image_id
                coco_prediction_json = coco_prediction.json
                coco_json.append(coco_prediction_json)

        time_start = time.time()
        # export prediction boxes
        if export_crop:
            crop_object_predictions(
                image=image,
                object_prediction_list=object_prediction_list,
                output_dir=str(crop_dir),
                file_name=filename_without_extension,
                export_format=visual_export_format,
            )
        # export prediction list as pickle
        if export_pickle:
            save_path = str(pickle_dir / (filename_without_extension + ".pickle"))
            save_pickle(data=object_prediction_list, save_path=save_path)
        # export visualization
        if export_visual:
            visualize_object_predictions(
                image,
                object_prediction_list=object_prediction_list,
                rect_th=visual_bbox_thickness,
                text_size=visual_text_size,
                text_th=visual_text_thickness,
                output_dir=str(visual_dir),
                file_name=filename_without_extension,
                export_format=visual_export_format,
            )
        time_end = time.time() - time_start
        durations_in_seconds["export_files"] = time_end

    # export coco results
    if coco_file_path:
        save_path = str(save_dir / "result.json")
        save_json(coco_json, save_path)

    # print prediction duration
    if verbose == 1:
        print(
            "Model loaded in",
            durations_in_seconds["model_load"],
            "seconds.",
        )
        print(
            "Slicing performed in",
            durations_in_seconds["slice"],
            "seconds.",
        )
        print(
            "Prediction performed in",
            durations_in_seconds["prediction"],
            "seconds.",
        )
        if export_visual:
            print(
                "Exporting performed in",
                durations_in_seconds["export_files"],
                "seconds.",
            )



# OBSS SAHI Tool
# Code written by Fatih C Akyon, 2020.

import os
import time
from typing import Dict, List

import cv2
import numpy as np
from tqdm import tqdm

from sahi.utils.coco import Coco, CocoAnnotation, CocoImage, create_coco_dict
from sahi.utils.cv import read_large_image
from sahi.utils.file import create_dir, load_json, save_json
from sahi.utils.shapely import (ShapelyAnnotation, get_bbox_from_shapely,
                                get_shapely_box, get_shapely_multipolygon)


def slice_coco_annotations_by_box(
    coco_annotation_list, box
) -> (List[CocoAnnotation], int):
    """
    Crops all annotations for a given grid. Returns processed annotation list.

    Args:
        coco_annotations: list of CocoAnnotation
            List of CocoAnnotation objects that belong to given COCO image
        box: list
            Points of slice box [x: int, y: int, width: int, height: int]
    Returns:
        sliced_annotations: list of CocoAnnotation
            Sliced list of CocoAnnotation objects that belong to given COCO image
    """

    sliced_annotations = []
    num_invalid_segmentation = 0
    for coco_annotation in coco_annotation_list:
        # calculate intersection polygon btw slice and annotation
        x, y, width, height = box
        shapely_box = get_shapely_box(x, y, width, height)

        shapely_annotation = coco_annotation._shapely_annotation

        invalid_segmentation = False
        if shapely_annotation.multipolygon.is_valid is True:
            intersection_shapely_annotation = shapely_annotation.get_intersection(
                shapely_box
            )
        else:
            try:
                shapely_annotation = shapely_annotation.get_buffered_shapely_annotation(
                    distance=3
                )
                intersection_shapely_annotation = shapely_annotation.get_intersection(
                    shapely_box
                )
            except:
                invalid_segmentation = True
                num_invalid_segmentation += 1

        # store intersection polygon if intersection area is greater than 0
        if (invalid_segmentation is False) and (
            intersection_shapely_annotation.area > 0
        ):
            new_annotation = CocoAnnotation.from_coco_segmentation(
                segmentation=intersection_shapely_annotation.to_coco_segmentation(),
                category_id=coco_annotation.category_id,
                category_name=coco_annotation.category_name,
            )

            sliced_annotations.append(new_annotation)

    return sliced_annotations, num_invalid_segmentation


class SlicedImage:
    def __init__(self, image, coco_image, starting_pixel):
        """
        image: np.array
            Sliced image.
        coco_image: CocoImage
            Coco styled image object that belong to sliced image.
        starting_pixel: list of list of int
            Starting pixel coordinates of the sliced image.
        """
        self.image = image
        self.coco_image = coco_image
        self.starting_pixel = starting_pixel


class SliceImageResult:
    def __init__(self, original_image_size=None, image_dir=None):
        """
        sliced_image_list: list of SlicedImage
        image_dir: str
            Directory of the sliced image exports.
        original_image_size: list of int
            Size of the unsliced original image in [height, width]
        """
        self._sliced_image_list = []
        self.original_image_height = original_image_size[0]
        self.original_image_width = original_image_size[1]
        self.image_dir = image_dir

    def add_sliced_image(self, sliced_image: SlicedImage):
        assert (
            type(sliced_image) == SlicedImage
        ), "sliced_image must be a SlicedImage instance"
        self._sliced_image_list.append(sliced_image)

    @property
    def sliced_image_list(self):
        return self._sliced_image_list

    @property
    def images(self):
        """
        Returns sliced images.

        Returns:
            images: a list of np.array
        """
        images = []
        for sliced_image in self._sliced_image_list:
            images.append(sliced_image.image)
        return images

    @property
    def coco_images(self):
        """
        Returns CocoImage representation of SliceImageResult.

        Returns:
            coco_images: a list of CocoImage
        """
        coco_images = []
        for sliced_image in self._sliced_image_list:
            coco_images.append(sliced_image.coco_image)
        return coco_images

    @property
    def starting_pixels(self):
        """
        Returns a list of starting pixels for each slice.

        Returns:
            starting_pixels: a list of starting pixel coords [x,y]
        """
        starting_pixels = []
        for sliced_image in self._sliced_image_list:
            starting_pixels.append(sliced_image.starting_pixel)
        return starting_pixels

    def __len__(self):
        return len(self._sliced_image_list)


def slice_image(
    image,
    coco_annotation_list=None,
    output_file_name: str = "",
    output_dir: str = "",
    slice_height: int = 256,
    slice_width: int = 256,
    max_allowed_zeros_ratio: float = 0.2,
    overlap_height_ratio: float = 0.2,
    overlap_width_ratio: float = 0.2,
    slice_sep: str = "_",
    out_ext: str = ".png",
    verbose: bool = False,
) -> (SliceImageResult, int):

    """
    Slice a large image into smaller windows. If output_file_name is given export sliced images.

    Args:
        image: str or np.ndarray
            Location of image or numpy image matrix to slice
        coco_annotation_list: list of CocoAnnotation
            List of CocoAnnotation objects that belong to given COCO image.
        output_file_name: str
            Root name of output files (coordinates will be appended to this)
        output_dir: str
            Output directory
        slice_height: int
            Height of each slice. Defaults to ``256``.
        slice_width: int
            Width of each slice. Defaults to ``256``.
        max_allowed_zeros_ratio: float
            Maximum fraction of window that is allowed to be zeros or null.
            Defaults to ``0.2``.
        overlap_height_ratio: float
            Fractional overlap in height of each window (e.g. an overlap of 0.2 for a window
            of size 256 yields an overlap of 51 pixels).
            Default to ``0.2``.
        overlap_width_ratio: float
            Fractional overlap in width of each window (e.g. an overlap of 0.2 for a window
            of size 256 yields an overlap of 51 pixels).
            Default to ``0.2``.
        slice_sep: str
            Character used to separate outname from coordinates in the saved
            windows. Defaults to ``|``
        out_ext: str
            Extension of saved images. Defaults to ``.png``.
        verbose: bool
            Switch to print relevant values to screen. Defaults to ``False``

    Returns:
        sliced_image_result: SliceImageResult:
                                sliced_image_list: list of SlicedImage
                                image_dir: str
                                    Directory of the sliced image exports.
                                original_image_size: list of int
                                    Size of the unsliced original image in [height, width]
        num_total_invalid_segmentation: int
            Number of invalid segmentation annotations.
    """

    # define verboseprint
    verboseprint = print if verbose else lambda *a, **k: None

    # create outdir if not present
    if output_dir:
        create_dir(output_dir)

    # read image if str image path is provided
    if isinstance(image, str):
        # read in image, cv2 fails on large files
        verboseprint("Read in image:", image)
        image0, use_cv2 = read_large_image(image)
    else:
        image0 = image
    verboseprint("image.shape:", image0.shape)

    if len(out_ext) == 0:
        ext = "." + image.split(".")[-1]
    else:
        ext = out_ext

    win_h, win_w = image0.shape[:2]

    # if slice sizes are large than image, pad the edges
    pad = 0
    if slice_height > win_h:
        pad = slice_height - win_h
    if slice_width > win_w:
        pad = max(pad, slice_width - win_w)
    # pad the edge of the image with black pixels
    if pad > 0:
        border_color = (0, 0, 0)
        image0 = cv2.copyMakeBorder(
            image0, 0, pad, 0, pad, cv2.BORDER_CONSTANT, value=border_color
        )

    win_size = slice_height * slice_width

    t0 = time.time()
    n_ims = 0
    n_ims_nonull = 0
    dx = int((1.0 - overlap_width_ratio) * slice_width)
    dy = int((1.0 - overlap_height_ratio) * slice_height)

    # init images and annotations lists
    sliced_image_result = SliceImageResult(
        original_image_size=[win_h, win_w], image_dir=output_dir
    )
    num_total_invalid_segmentation = 0

    # iterate over slices
    for y0 in range(0, image0.shape[0], dy):  # slice_height):
        for x0 in range(0, image0.shape[1], dx):  # slice_width):
            n_ims += 1

            if (n_ims % 50) == 0:
                verboseprint(n_ims)

            # make sure we don't have a tiny image on the edge
            if y0 + slice_height > image0.shape[0]:
                y = image0.shape[0] - slice_height
            else:
                y = y0
            if x0 + slice_width > image0.shape[1]:
                x = image0.shape[1] - slice_width
            else:
                x = x0

            # extract image
            window_c = image0[y : y + slice_height, x : x + slice_width]

            # process annotations if coco_annotations is given
            if coco_annotation_list:
                slice_box = [x, y, slice_width, slice_height]
                (
                    sliced_coco_annotation_list,
                    num_invalid_segmentation,
                ) = slice_coco_annotations_by_box(coco_annotation_list, box=slice_box)
                num_total_invalid_segmentation = (
                    num_total_invalid_segmentation + num_invalid_segmentation
                )

            # get black and white image
            window = cv2.cvtColor(window_c, cv2.COLOR_RGB2GRAY)

            # find threshold that's not black
            # https://opencv-python-tutroals.readthedocs.org/en/latest/py_tutorials/py_imgproc/py_thresholding/py_thresholding.html?highlight=threshold
            ret, thresh1 = cv2.threshold(window, 2, 255, cv2.THRESH_BINARY)
            non_zero_counts = cv2.countNonZero(thresh1)
            zero_counts = win_size - non_zero_counts
            zero_frac = float(zero_counts) / win_size
            # skip if image is mostly empty
            if zero_frac >= max_allowed_zeros_ratio:
                verboseprint("Zero frac too high at:", zero_frac)
                continue
            else:
                # save if out_name is given
                if output_file_name and output_dir:
                    outpath = os.path.join(
                        output_dir,
                        output_file_name
                        + slice_sep
                        + str(y)
                        + "_"
                        + str(x)
                        + "_"
                        + str(slice_height)
                        + "_"
                        + str(slice_width)
                        + "_"
                        + str(pad)
                        + "_"
                        + str(win_w)
                        + "_"
                        + str(win_h)
                        + ext,
                    )

                    verboseprint("outpath:", outpath)
                    # if large image, convert to bgr prior to saving
                    if not use_cv2:
                        try:
                            import skimage.io
                        except ImportError:
                            raise ImportError(
                                'Please run "pip install -U scikit-image" '
                                'to install scikit-image first for large image handling.')
                        skimage.io.imsave(outpath, window_c)
                    else:
                        window_c = cv2.cvtColor(window_c, cv2.COLOR_RGB2BGR)
                        cv2.imwrite(outpath, window_c)
                    n_ims_nonull += 1

                    file_name = outpath.split(output_dir)[-1].replace(os.sep, "")
                else:
                    file_name = ""

                # create coco image
                coco_image = CocoImage(
                    file_name=file_name, height=slice_height, width=slice_width
                )

                # append coco annotations (if present) to coco image
                if coco_annotation_list:
                    for coco_annotation in sliced_coco_annotation_list:
                        coco_image.add_annotation(coco_annotation)

                # create sliced image and append to sliced_image_result
                sliced_image = SlicedImage(
                    image=window_c, coco_image=coco_image, starting_pixel=[x, y]
                )
                sliced_image_result.add_sliced_image(sliced_image)

    verboseprint(
        "Num slices:",
        n_ims,
        "Num non-null slices:",
        n_ims_nonull,
        "slice_height",
        slice_height,
        "slice_width",
        slice_width,
    )
    verboseprint("Time to slice", image, time.time() - t0, "seconds")

    return (
        sliced_image_result,
        num_total_invalid_segmentation,
    )


def slice_coco(
    coco_annotation_file_path: str,
    image_dir: str,
    output_coco_annotation_file_name: str = "",
    output_dir: str = "",
    ignore_negative_samples: bool = True,
    slice_height: int = 256,
    slice_width: int = 256,
    max_allowed_zeros_ratio: float = 0.2,
    overlap_height_ratio: float = 0.2,
    overlap_width_ratio: float = 0.2,
    slice_sep: str = "_",
    out_ext: str = ".png",
    verbose: bool = False,
):

    """
    Slice large images given in a directory, into smaller windows. If out_name is given export sliced images and coco file.

    Args:
        coco_annotation_file_path: str
            Location of the coco annotation file
        image_dir: str
            Base diectory for the images
        output_coco_annotation_file_name : str
            Root name of the exported coco datatset file
        output_dir: str
            Output directory
        ignore_negative_samples: bool
            If True, images without annotations are ignored. Defaults to True.
        slice_height: int
            Height of each slice.  Defaults to ``256``.
        slice_width: int
            Width of each slice.  Defaults to ``256``.
        max_allowed_zeros_ratio: float
            Maximum fraction of window that is allowed to be zeros or null.
            Defaults to ``0.2``.
        overlap_height_ratio: float
            Fractional overlap in height of each window (e.g. an overlap of 0.2 for a window
            of size 256 yields an overlap of 51 pixels).
            Default to ``0.2``.
        overlap_width_ratio: float
            Fractional overlap in width of each window (e.g. an overlap of 0.2 for a window
            of size 256 yields an overlap of 51 pixels).
            Default to ``0.2``.
        slice_sep: str
            Character used to separate outname from coordinates in the saved
            windows. Defaults to ``|``
        out_ext: str
            Extension of saved images. Defaults to ``.png``.
        verbose: bool
            Switch to print relevant values to screen. Defaults to ``False``

    Returns:
        coco_dict: dict
            COCO dict for sliced images and annotations
        save_path: str
            Path to the saved coco file
    """
    # define verboseprint
    verboseprint = print if verbose else lambda *a, **k: None

    # read coco file
    coco_dict = load_json(coco_annotation_file_path)
    # create coco_utils.Coco object
    coco = Coco.from_coco_dict_or_path(coco_dict)
    # init sliced coco_utils.CocoImage list
    sliced_coco_images = []

    num_total_invalid_segmentation = 0

    # iterate over images and slice
    for coco_image in tqdm(coco.images):
        # get image path
        image_path = os.path.join(image_dir, coco_image.file_name)
        # get coco_utils.CocoAnnotation list corresponding to selected coco_utils.CocoImage
        coco_annotation_list = coco_image.annotations
        # slice image
        slice_image_result, num_invalid_segmentation = slice_image(
            image=image_path,
            coco_annotation_list=coco_annotation_list,
            output_file_name=os.path.basename(coco_image.file_name),
            output_dir=output_dir,
            slice_height=slice_height,
            slice_width=slice_width,
            max_allowed_zeros_ratio=max_allowed_zeros_ratio,
            overlap_height_ratio=overlap_height_ratio,
            overlap_width_ratio=overlap_width_ratio,
            slice_sep="_",
            out_ext=".png",
            verbose=False,
        )
        num_total_invalid_segmentation = (
            num_total_invalid_segmentation + num_invalid_segmentation
        )
        # append slice outputs
        sliced_coco_images.extend(slice_image_result.coco_images)

    # create and save coco dict
    coco_dict = create_coco_dict(
        sliced_coco_images,
        coco_dict["categories"],
        ignore_negative_samples=ignore_negative_samples,
    )
    save_path = ""
    if output_coco_annotation_file_name and output_dir:
        save_path = os.path.join(
            output_dir, output_coco_annotation_file_name + "_coco.json"
        )
        save_json(coco_dict, save_path)
    verboseprint("There are", num_total_invalid_segmentation, "invalid segmentations")

    return coco_dict, save_path



# OBSS SAHI Tool
# Code written by Fatih C Akyon, 2020.

import copy

import numpy as np

from sahi.annotation import ObjectAnnotation
from sahi.utils.torch import to_float_tensor, torch_stack


class PredictionScore:
    def __init__(self, score: float):
        """
        Arguments:
            score: prediction score between 0 and 1
        """
        # if score is a numpy object, convert it to python variable
        if type(score).__module__ == "numpy":
            score = copy.deepcopy(score).tolist()
        # set score
        self.score = score

    def is_greater_than_threshold(self, threshold):
        """
        Check if score is greater than threshold
        """
        return self.score > threshold

    def __repr__(self):
        return f"PredictionScore: <score: {self.score}>"


class ObjectPrediction(ObjectAnnotation):
    """
    Class for handling detection model predictions.
    """

    def __init__(
        self,
        bbox: list,
        score: float,
        category_id: int,
        category_name=None,
        bool_mask=None,
        shift_amount: list = [0, 0],
        full_shape: list = [1024, 1024],
    ):
        """
        Creates ObjectPrediction from bbox, score, category_id, category_name, bool_mask.

        Arguments:
            bbox: list
                [minx, miny, maxx, maxy]
            score: float
                Prediction score between 0 and 1
            category_id: int
                ID of the object category
            category_name: str
                Name of the object category
            bool_mask: np.ndarray
                2D boolean mask array. Should be None if model doesn't output segmentation mask.
            shift_amount: list
                To shift the box and mask predictions from sliced image
                to full sized image, should be in the form of [shift_x, shift_y]
            full_shape: list
                Size of the full image after shifting, should be in
                the form of [height, width]
        """
        self.score = PredictionScore(score)
        super().__init__(
            bbox=bbox,
            category_id=category_id,
            bool_mask=bool_mask,
            category_name=category_name,
            shift_amount=shift_amount,
            full_shape=full_shape,
        )

    def get_shifted_object_prediction(self):
        """
        Returns shifted version ObjectPrediction.
        Shifts bbox and mask coords.
        Used for mapping sliced predictions over full image.
        """
        if self.mask:
            return ObjectPrediction(
                bbox=self.bbox.get_shifted_box().to_voc_bbox(),
                category_id=self.category.id,
                score=self.score.score,
                bool_mask=self.mask.get_shifted_mask().bool_mask,
                category_name=self.category.name,
                shift_amount=[0, 0],
                full_shape=self.mask.get_shifted_mask().full_shape,
            )
        else:
            return ObjectPrediction(
                bbox=self.bbox.get_shifted_box().to_voc_bbox(),
                category_id=self.category.id,
                score=self.score.score,
                bool_mask=None,
                category_name=self.category.name,
                shift_amount=[0, 0],
                full_shape=None,
            )

    def __repr__(self):
        return f"""ObjectPrediction<
    bbox: {self.bbox},
    mask: {self.mask},
    score: {self.score},
    category: {self.category}>"""


class PredictionInput:
    def __init__(
        self,
        image_list,
        shift_amount_list=None,
        full_shape=None,
    ):
        """
        Arguments:
            image_list: list of images to be predicted
            shift_amount_list: To shift the box and mask predictions from sliced image to full sized image, should be in the form of [shift_x, shift_y]
            full_shape: Size of the full image after shifting, should be in the form of [height, width]

        image_list and shift_amount_list should have same length
        """
        self.image_list = image_list
        image_tensor_list = []
        for image in image_list:
            # normalize image
            image = image / np.max(image)
            # convert numpy image to tensor
            image_tensor_list.append(to_float_tensor(image))
        # create batch tensor
        batch_image_tensor = torch_stack(tuple(image_tensor_list), dim=0)
        self.batch_image_tensor = batch_image_tensor
        # set other properties
        if shift_amount_list:
            self.shift_amount_list = shift_amount_list
        else:
            self.shift_amount_list = [[0, 0] for ind in range(len(image_list))]
        self.full_shape = full_shape



# OBSS SAHI Tool
# Code written by Fatih C Akyon, 2020.

import copy
import os
import random
import time

import cv2
import numpy as np
from sahi.utils.file import create_dir


def crop_object_predictions(
    image: np.array,
    object_prediction_list,
    output_dir: str = "",
    file_name: str = "prediction_visual",
    export_format: str = "png",
):
    """
    Crops bounding boxes over the source image and exports it to output folder.
    Arguments:
        object_predictions: a list of prediction.ObjectPrediction
        output_dir: directory for resulting visualization to be exported
        file_name: exported file will be saved as: output_dir+file_name+".png"
        export_format: can be specified as 'jpg' or 'png'
    """
    # create output folder if not present
    create_dir(output_dir)
    # add bbox and mask to image if present
    for ind, object_prediction in enumerate(object_prediction_list):
        # deepcopy object_prediction_list so that original is not altered
        object_prediction = object_prediction.deepcopy()
        bbox = object_prediction.bbox.to_voc_bbox()
        category_id = object_prediction.category.id
        # crop detections
        # deepcopy crops so that original is not altered
        cropped_img = copy.deepcopy(
            image[
                int(bbox[1]) : int(bbox[3]),
                int(bbox[0]) : int(bbox[2]),
                :,
            ]
        )
        save_path = os.path.join(
            output_dir,
            file_name
            + "_box"
            + str(ind)
            + "_class"
            + str(category_id)
            + "."
            + export_format,
        )
        cv2.imwrite(save_path, cv2.cvtColor(cropped_img, cv2.COLOR_RGB2BGR))


def convert_image_to(read_path, extension="jpg", grayscale=False):
    """
    Reads image from path and saves as given extension.
    """
    image = cv2.imread(read_path)
    pre, ext = os.path.splitext(read_path)
    if grayscale:
        image = cv2.cvtColor(image, cv2.COLOR_BGR2GRAY)
        pre = pre + "_gray"
    save_path = pre + "." + extension
    cv2.imwrite(save_path, image)


def read_large_image(image_path):
    use_cv2 = True
    # read image, cv2 fails on large files
    try:
        # convert to rgb (cv2 reads in bgr)
        img_cv2 = cv2.imread(image_path, 1)
        image0 = cv2.cvtColor(img_cv2, cv2.COLOR_BGR2RGB)
    except:
        try:
            import skimage.io
        except ImportError:
            raise ImportError(
                'Please run "pip install -U scikit-image" '
                'to install scikit-image first for large image handling.')
        image0 = skimage.io.imread(image_path, as_grey=False).astype(np.uint8)  # [::-1]
        use_cv2 = False
    return image0, use_cv2


def read_image(image_path):
    """
    Loads image as numpy array from given path.
    """
    # read image
    image = cv2.imread(image_path)
    image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)
    # return image
    return image


def select_random_color():
    """
    Selects random color.
    """
    colors = [
        [0, 255, 0],
        [0, 0, 255],
        [255, 0, 0],
        [0, 255, 255],
        [255, 255, 0],
        [255, 0, 255],
        [80, 70, 180],
        [250, 80, 190],
        [245, 145, 50],
        [70, 150, 250],
        [50, 190, 190],
    ]
    return colors[random.randrange(0, 10)]


def apply_color_mask(image: np.array, color: tuple):
    """
    Applies color mask to given input image.
    """
    r = np.zeros_like(image).astype(np.uint8)
    g = np.zeros_like(image).astype(np.uint8)
    b = np.zeros_like(image).astype(np.uint8)

    (r[image == 1], g[image == 1], b[image == 1]) = color
    colored_mask = np.stack([r, g, b], axis=2)
    return colored_mask


def visualize_prediction(
    image: np.array,
    boxes,
    classes,
    masks=None,
    rect_th: float = 3,
    text_size: float = 3,
    text_th: float = 3,
    color: tuple = (0, 0, 0),
    output_dir: str = "",
    file_name: str = "prediction_visual",
):
    """
    Visualizes prediction classes, bounding boxes over the source image
    and exports it to output folder.
    """
    elapsed_time = time.time()
    # deepcopy image so that original is not altered
    image = copy.deepcopy(image)
    # select random color if not specified
    if color == (0, 0, 0):
        color = select_random_color()
    # add bbox and mask to image if present
    for i in range(len(boxes)):
        # deepcopy boxso that original is not altered
        box = copy.deepcopy(boxes[i])
        class_ = classes[i]
        # visualize masks if present
        if masks is not None:
            # deepcopy mask so that original is not altered
            mask = copy.deepcopy(masks[i])
            # draw mask
            rgb_mask = apply_color_mask(np.squeeze(mask), color)
            image = cv2.addWeighted(image, 1, rgb_mask, 0.7, 0)
        # visualize boxes
        cv2.rectangle(
            image,
            tuple(box[0:2]),
            tuple(box[2:4]),
            color=color,
            thickness=rect_th,
        )
        # arange bounding box text location
        if box[1] - 10 > 10:
            box[1] -= 10
        else:
            box[1] += 10
        # add bounding box text
        cv2.putText(
            image,
            class_,
            tuple(box[0:2]),
            cv2.FONT_HERSHEY_SIMPLEX,
            text_size,
            color,
            thickness=text_th,
        )
    if output_dir != "":
        # create output folder if not present
        create_dir(output_dir)
        # save inference result
        save_path = os.path.join(output_dir, file_name + ".png")
        cv2.imwrite(save_path, cv2.cvtColor(image, cv2.COLOR_RGB2BGR))
    elapsed_time = time.time() - elapsed_time
    return {"image": image, "elapsed_time": elapsed_time}


def visualize_object_predictions(
    image: np.array,
    object_prediction_list,
    rect_th: float = 1,
    text_size: float = 0.3,
    text_th: float = 1,
    color: tuple = (0, 0, 0),
    output_dir: str = "",
    file_name: str = "prediction_visual",
    export_format: str = "png",
):
    """
    Visualizes prediction category names, bounding boxes over the source image
    and exports it to output folder.
    Arguments:
        object_prediction_list: a list of prediction.ObjectPrediction
        rect_th: rectangle thickness
        text_size: size of the category name over box
        text_th: text thickness
        color: annotation color in the form: (0, 255, 0)
        output_dir: directory for resulting visualization to be exported
        file_name: exported file will be saved as: output_dir+file_name+".png"
        export_format: can be specified as 'jpg' or 'png'
    """
    elapsed_time = time.time()
    # deepcopy image so that original is not altered
    image = copy.deepcopy(image)
    # select random color if not specified
    if color == (0, 0, 0):
        color = select_random_color()
    # add bbox and mask to image if present
    for object_prediction in object_prediction_list:
        # deepcopy object_prediction_list so that original is not altered
        object_prediction = object_prediction.deepcopy()

        bbox = object_prediction.bbox.to_voc_bbox()
        category_name = object_prediction.category.name
        score = object_prediction.score.score

        # visualize masks if present
        if object_prediction.mask is not None:
            # deepcopy mask so that original is not altered
            mask = object_prediction.mask.bool_mask
            # draw mask
            rgb_mask = apply_color_mask(mask, color)
            image = cv2.addWeighted(image, 1, rgb_mask, 0.4, 0)
        # visualize boxes
        cv2.rectangle(
            image,
            tuple(bbox[0:2]),
            tuple(bbox[2:4]),
            color=color,
            thickness=rect_th,
        )
        # arange bounding box text location
        if bbox[1] - 5 > 5:
            bbox[1] -= 5
        else:
            bbox[1] += 5
        # add bounding box text
        label = "%s %.2f" % (category_name, score)
        cv2.putText(
            image,
            label,
            tuple(bbox[0:2]),
            cv2.FONT_HERSHEY_SIMPLEX,
            text_size,
            color,
            thickness=text_th,
        )
    if output_dir != "":
        # create output folder if not present
        create_dir(output_dir)
        # save inference result
        save_path = os.path.join(output_dir, file_name + "." + export_format)
        cv2.imwrite(save_path, cv2.cvtColor(image, cv2.COLOR_RGB2BGR))
    elapsed_time = time.time() - elapsed_time
    return {"image": image, "elapsed_time": elapsed_time}


def get_coco_segmentation_from_bool_mask(bool_mask):
    """
    Convert boolean mask to coco segmentation format
    [
        [x1, y1, x2, y2, x3, y3, ...],
        [x1, y1, x2, y2, x3, y3, ...],
        ...
    ]
    """
    # Generate polygons from mask
    mask = np.squeeze(bool_mask)
    mask = mask.astype(np.uint8)
    mask = cv2.copyMakeBorder(mask, 1, 1, 1, 1, cv2.BORDER_CONSTANT, value=0)
    polygons = cv2.findContours(
        mask, cv2.RETR_LIST, cv2.CHAIN_APPROX_SIMPLE, offset=(-1, -1)
    )
    polygons = polygons[0] if len(polygons) == 2 else polygons[1]
    # Convert polygon to coco segmentation
    coco_segmentation = [polygon.flatten().tolist() for polygon in polygons]
    return coco_segmentation


def get_bool_mask_from_coco_segmentation(coco_segmentation, width, height):
    """
    Convert coco segmentation to 2D boolean mask of given height and width
    """
    size = [height, width]
    points = [
        np.array(point).reshape(-1, 2).round().astype(int)
        for point in coco_segmentation
    ]
    bool_mask = np.zeros(size)
    bool_mask = cv2.fillPoly(bool_mask, points, 1)
    bool_mask.astype(bool)
    return bool_mask


def get_bbox_from_bool_mask(bool_mask):
    """
    Generate voc bbox ([xmin, ymin, xmax, ymax]) from given bool_mask (2D np.ndarray)
    """
    rows = np.any(bool_mask, axis=1)
    cols = np.any(bool_mask, axis=0)

    if not np.any(rows) or not np.any(cols):
        return []

    ymin, ymax = np.where(rows)[0][[0, -1]]
    xmin, xmax = np.where(cols)[0][[0, -1]]
    # width = xmax - xmin
    # height = ymax - ymin

    return [xmin, ymin, xmax, ymax]


def normalize_numpy_image(image: np.ndarray):
    """
    Normalizes numpy image
    """
    return image / np.max(image)


def ipython_display(image):
    """
    Displays numpy image in notebook.

    If input image is in range 0..1, please first multiply img by 255
    Assumes image is ndarray of shape [height, width, channels] where channels can be 1, 3 or 4
    """
    import IPython

    image = cv2.cvtColor(image, cv2.COLOR_RGB2BGR)
    _, ret = cv2.imencode(".png", image)
    i = IPython.display.Image(data=ret)
    IPython.display.display(i)
